@@ -21,69 +21,124 @@
 
 #include "config.h"
 
-#include <stdio.h>
+#include <assert.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
+#include <stdio.h>
 #include <errno.h>
-#include <assert.h>
+
+#include <quvi/quvi.h>
+#include <quvi/llst.h>
 
 #include <curl/curl.h>
 
 #include "platform.h"
-
-#include "quvi/quvi.h"
-#include "quvi/llst.h"
-
 #include "cmdline.h"
 
 #define _free(p) \
     do { if (p) { free(p); p=0; } } while (0)
 
+static const char *config_fname = ".quvirc";
+
 /* strepl.c */
 extern char *strepl(const char *s, const char *what, const char *with);
 
-static int verbose_flag = 1;
+/*@null@*/
 static quvi_t quvi = NULL;
+/*@null@*/
 static CURL *curl = NULL;
 
 typedef struct gengetopt_args_info *opts_t;
+/*@null@*/
 static opts_t opts = NULL;
 
-static quvi_llst_node_t inputs = NULL;
+/*@null@*/
+static quvi_llst_node_t input = NULL;
 
-/* prints to std(e)rr. */
+/* Check whether a message should be printed. */
+static int check_verbosity(int threshold)
+{
+  if (opts->verbosity_arg >= threshold)
+    return (0);
+  return (1);
+}
+
+/* Prints to std(e)rr. 'mute' disables. */
 static void spew_e(const char *fmt, ...)
 {
   va_list ap;
-  va_start(ap, fmt);
-  vfprintf(stderr, fmt, ap);
-  va_end(ap);
-}
-
-/* respects (q)uiet, prints to std(e)rr. */
-static void spew_qe(const char *fmt, ...)
-{
-  va_list ap;
-  if (!verbose_flag)
+  if (check_verbosity(verbosity_arg_mute) == 0)
     return;
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   va_end(ap);
 }
 
-/* glorified printf. */
+/* Prints to std(e)rr. 'quiet' disables. */
+static void spew_qe(const char *fmt, ...)
+{
+  va_list ap;
+  if (check_verbosity(verbosity_arg_quiet) == 0)
+    return;
+  va_start(ap, fmt);
+  vfprintf(stderr, fmt, ap);
+  va_end(ap);
+}
+
+/* Glorified printf. 'mute' disables. */
 static void spew(const char *fmt, ...)
 {
   va_list ap;
+  if (check_verbosity(verbosity_arg_mute) == 0)
+    return;
   va_start(ap, fmt);
   vfprintf(stdout, fmt, ap);
   va_end(ap);
 }
 
+static void dump_error_json(quvi_t quvi, QUVIcode rc)
+{
+  spew_e("{\n"
+         "  \"error\": [\n"
+         "    {\n"
+         "      \"message\": \"%s\"\n"
+         "    }\n"
+         "  ]\n"
+         "}\n", quvi_strerror(quvi, rc));
+}
+
+static void dump_error_xml(quvi_t quvi, QUVIcode rc)
+{
+  spew_e("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+         "<error>\n"
+         "  <message>%s</message>\n"
+         "</error>\n",
+         quvi_strerror(quvi, rc));
+}
+
 static void dump_error(quvi_t quvi, QUVIcode rc)
 {
-  fprintf(stderr, "error: %s\n", quvi_strerror(quvi, rc));
+  switch (opts->export_level_arg)
+    {
+    case export_level_arg_media:
+    case export_level__NULL:
+    default:
+      spew_e("error: %s\n", quvi_strerror(quvi, rc));
+      break;
+    case export_level_arg_PLUS_errors:
+      switch (opts->export_format_arg)
+        {
+        case export_format_arg_json:
+        case export_format__NULL:
+        default:
+          dump_error_json(quvi, rc);
+          break;
+        case export_format_arg_xml:
+          dump_error_xml(quvi, rc);
+          break;
+        }
+    }
 }
 
 static void handle_resolve_status(quvi_word type)
@@ -153,22 +208,30 @@ static int status_callback(long param, void *data)
   return (0);
 }
 
-static size_t write_callback(void *p, size_t size, size_t nmemb,
-                             void *data)
+static void cleanup()
 {
-  size_t r = quvi_write_callback_default(p, size, nmemb, data);
-  /* Could do something useful here. */
-#ifdef _0
-  puts(__func__);
-#endif
-  return r;
+  if (input)
+    quvi_llst_free(&input);
+
+  if (quvi)
+    quvi_close(&quvi);
+
+  if (opts)
+    {
+      cmdline_parser_free(opts);
+      _free(opts);
+    }
+
+  assert(input == NULL);
+  assert(quvi == NULL);
+  assert(opts == NULL);
 }
 
 /* Divided into smaller blocks. Otherwise -pedantic objects. */
 
 #define LICENSE_1 \
 "/* quvi\n" \
-" * Copyright (C) 2009,2010,2011  Toni Gundogdu <legatvs@gmail.com>\n"
+" * Copyright (C) 2009-2011  Toni Gundogdu <legatvs@gmail.com>\n"
 
 #define LICENSE_2 \
 " * This library is free software; you can redistribute it and/or\n" \
@@ -192,7 +255,8 @@ static void license()
 {
   printf("%s *\n%s *\n%s *\n%s\n",
          LICENSE_1, LICENSE_2, LICENSE_3, LICENSE_4);
-  exit(0);
+  cleanup();
+  exit (0);
 }
 
 #undef LICENSE_4
@@ -200,10 +264,20 @@ static void license()
 #undef LICENSE_2
 #undef LICENSE_1
 
-static void version()
+static void print_version()
 {
-  printf("quvi version %s\n", quvi_version(QUVI_VERSION_LONG));
-  exit(0);
+  static const char version[] =
+#ifdef VN
+    VN
+#else
+    PACKAGE_VERSION
+#endif
+    " for " CANONICAL_TARGET;
+  printf("quvi %s\n  libquvi %s\n  libquvi-scripts %s\n",
+         version, quvi_version(QUVI_VERSION_LONG),
+         quvi_version(QUVI_VERSION_SCRIPTS));
+  cleanup();
+  exit (0);
 }
 
 static void dump_host(char *domain, char *formats)
@@ -216,12 +290,10 @@ static void dump_host(char *domain, char *formats)
 /* Wraps quvi_supported_ident. */
 static void supported(quvi_t quvi)
 {
-#ifndef _QUVI_SUPPORTED
   quvi_ident_t ident;
+  unsigned int i;
   char *formats;
-#endif
   QUVIcode rc;
-  int i;
 
   rc = QUVI_NOSUPPORT;
 
@@ -230,101 +302,32 @@ static void supported(quvi_t quvi)
       rc = quvi_supported_ident(quvi, (char*)opts->inputs[i], &ident);
       if (rc == QUVI_OK)
         {
-          quvi_ident_getprop(ident, QUVI_IDENT_PROPERTY_FORMATS, &formats);
+          quvi_ident_getprop(ident,
+                             QUVI_IDENT_PROPERTY_FORMATS, &formats);
           spew("%10s : %s\n", formats, (char *)opts->inputs[i]);
           quvi_supported_ident_close(&ident);
         }
       else
         dump_error(quvi,rc);
     }
-
-  exit(rc);
-}
-
-static const char depr_msg[] =
-  "Warning:\n"
-  "   '--format list' is deprecated and will be removed in\n"
-  "   quvi 0.2.20. Use --query-formats instead.";
-
-static void format_help(quvi_t quvi)
-{
-  int quit = 0;
-
-  if (strcmp(opts->format_arg, "help") == 0)
-    {
-      printf(
-        "Usage:\n"
-        "   --format arg                get format arg of media\n"
-        "   --format list               print domains with formats\n"
-        "   --format list arg           match arg to supported domain names\n"
-        "Examples:\n"
-        "   --format list youtube       print youtube formats\n"
-        "   --format fmt34_360p         get format fmt34_360p of media\n"
-        "%s\n",
-        depr_msg);
-      quit = 1;
-    }
-
-  else if (strcmp(opts->format_arg, "list") == 0)
-    {
-      int done = 0;
-      char *d, *f;
-
-      while(!done)
-        {
-          const int rc =
-            quvi_next_supported_website(quvi, &d, &f);
-
-          switch (rc)
-            {
-
-            case QUVI_OK:
-            {
-              int print = 1;
-
-              /* -f list <pattern> */
-              if (opts->inputs_num > 0)
-                print = strstr(d, (char *)opts->inputs[0]) != 0;
-
-              /* -f list */
-              if (print)
-                printf("%s:\n  %s\n\n", d, f);
-
-              quvi_free(d);
-              quvi_free(f);
-            }
-            break;
-
-            case QUVI_LAST:
-              done = 1;
-              break;
-
-            default:
-              spew_e("%s\n", quvi_strerror(quvi, rc));
-              break;
-            }
-        }
-      spew_qe("%s\n", depr_msg);
-      quit = 1;
-    }
-
-  if (quit)
-    exit(0);
+  cleanup();
+  exit (rc);
 }
 
 /* Query which formats are available for the URL */
 static void query_formats(quvi_t quvi)
 {
+  unsigned int i;
   QUVIcode rc;
-  int i;
 
-  if (opts->inputs_num < 1)
+  if (opts->inputs_num <1)
     {
       spew_qe("error: no input URLs\n");
+      cleanup();
       exit (QUVI_INVARG);
     }
 
-  for (i=0; i<opts->inputs_num; ++i)
+  for (i=0,rc=0; i<opts->inputs_num; ++i)
     {
       char *formats = NULL;
 
@@ -337,8 +340,8 @@ static void query_formats(quvi_t quvi)
       else
         dump_error(quvi,rc);
     }
-
-  exit(rc);
+  cleanup();
+  exit (rc);
 }
 
 /* dumps all supported hosts to stdout. */
@@ -346,10 +349,10 @@ static void support(quvi_t quvi)
 {
   int done = 0;
 
-  if (opts->inputs_num > 0)
+  if (opts->inputs_num >0)
     supported(quvi);
 
-  while (!done)
+  while (done == 0)
     {
       char *domain, *formats;
       QUVIcode rc;
@@ -369,22 +372,48 @@ static void support(quvi_t quvi)
           break;
         }
     }
-
-  exit(0);
+  cleanup();
+  exit (0);
 }
 
-static void invoke_exec(quvi_media_t media)
+static char * shell_escape( char *str )
 {
-  char *cmd, *media_url,  *q_media_url;
+  size_t len = strlen( str );
+  int i;
+  str = realloc( str, 2 * len + 1 );
+  if ( !str ) abort();
+  for ( i = len - 1; i >= 0; --i )
+    {
+      str[ 2 * i + 1 ] = str[ i ];
+      str[ 2 * i ]     = '\\';
+    }
+  str[ 2 * len ] = 0;
+  return str;
+}
+
+static void invoke_exec(quvi_media_t media, const char *exec_arg)
+{
+  char *cmd, *media_url, *q_media_url;
+  char *page_title, *q_page_title, *t;
   int rc;
 
+  quvi_getprop(media, QUVIPROP_PAGETITLE, &page_title);
+  t = strdup(page_title);
+  t = shell_escape(t);
+  asprintf(&q_page_title, "%s", t);
+  _free(t);
+
   quvi_getprop(media, QUVIPROP_MEDIAURL, &media_url);
+  t = strdup(media_url);
+  t = shell_escape(t);
+  asprintf(&q_media_url, "%s", t);
+  _free(t);
 
-  asprintf(&q_media_url, "\"%s\"", media_url);
-
-  cmd = strdup(opts->exec_arg);
+  cmd = strdup(exec_arg);
+  cmd = strepl(cmd, "%t", q_page_title);
   cmd = strepl(cmd, "%u", q_media_url);
 
+  _free(q_page_title);
   _free(q_media_url);
 
   rc = system(cmd);
@@ -400,7 +429,6 @@ static void invoke_exec(quvi_media_t media)
       spew_e("error: child exited with: %d\n", rc >> 8);
       break;
     }
-
   _free(cmd);
 }
 
@@ -420,51 +448,39 @@ static void dump_media_url_xml(parsed_url_t p, int i)
 
   spew("   <link id=\"%d\">\n", i);
 
-  if (p->content_length)
+  if (p->content_length >0)
     spew("       <length_bytes>%.0f</length_bytes>\n", p->content_length);
 
-  if (strlen(p->content_type))
+  if (strlen(p->content_type) >0)
     spew("       <content_type>%s</content_type>\n", p->content_type);
 
-  if (strlen(p->file_suffix))
+  if (strlen(p->file_suffix) >0)
     spew("       <file_suffix>%s</file_suffix>\n", p->file_suffix);
 
   spew("       <url>%s</url>\n"
        "   </link>\n",
-       media_url ? media_url : p->media_url);
+       (media_url != NULL)
+       ? media_url
+       : p->media_url);
 
   _free(media_url);
 }
 
-static void dump_media_url_old(parsed_url_t p, int i)
-{
-  spew("link %02d  : %s\n", i, p->media_url);
-
-  if (p->content_length)
-    spew(":: length: %.0f\n", p->content_length);
-
-  if (strlen(p->file_suffix))
-    spew(":: suffix: %s\n", p->file_suffix);
-
-  if (strlen(p->content_type))
-    spew(":: content-type: %s\n", p->content_type);
-}
-
 static void dump_media_url_json(parsed_url_t p, int i, int prepend_newln)
 {
-  if (prepend_newln)
+  if (prepend_newln != 0)
     spew(",\n");
 
   spew("    {\n"
        "      \"id\": \"%d\",\n", i);
 
-  if (p->content_length)
+  if (p->content_length >0)
     spew("      \"length_bytes\": \"%.0f\",\n", p->content_length);
 
-  if (strlen(p->content_type))
+  if (strlen(p->content_type) >0)
     spew("      \"content_type\": \"%s\",\n", p->content_type);
 
-  if (strlen(p->file_suffix))
+  if (strlen(p->file_suffix) >0)
     spew("      \"file_suffix\": \"%s\",\n", p->file_suffix);
 
   spew("      \"url\": \"%s\"\n"
@@ -480,26 +496,23 @@ static void dump_media_urls(quvi_media_t media)
       struct parsed_url_s p;
 
       memset(&p, 0, sizeof(&p));
-
       quvi_getprop(media, QUVIPROP_MEDIAURL, &p.media_url);
       quvi_getprop(media, QUVIPROP_MEDIACONTENTTYPE, &p.content_type);
       quvi_getprop(media, QUVIPROP_MEDIACONTENTLENGTH, &p.content_length);
       quvi_getprop(media, QUVIPROP_FILESUFFIX, &p.file_suffix);
 
-      if (opts->xml_given)
+      if (opts->xml_given == 1)
         dump_media_url_xml(&p,i);
-      else if (opts->old_given)
-        dump_media_url_old(&p,i);
       else
         {
-          dump_media_url_json(&p, i, i>1);
+          dump_media_url_json(&p, i, (int)(i>1));
           json_flag = 1;
         }
       ++i;
     }
   while (quvi_next_media_url(media) == QUVI_OK);
 
-  if (json_flag)
+  if (json_flag == 1)
     spew("\n");
 }
 
@@ -535,52 +548,28 @@ static void dump_media_xml(parsed_t p)
        p->page_title,
        e_page_url ? e_page_url : "");
 
-  if (strlen(p->start_time))
+  if (strlen(p->start_time) >0)
     spew("   <start_time>%s</start_time>\n", p->start_time);
 
-  if (e_thumb_url && strlen(e_thumb_url))
+  if (e_thumb_url != NULL && strlen(e_thumb_url) >0)
     spew("   <thumbnail_url>%s</thumbnail_url>\n", e_thumb_url);
 
-  if (p->duration)
+  if (p->duration >0)
     spew("   <duration>%.0f</duration>\n", p->duration);
 
-  if (e_page_url)
+  if (e_page_url != NULL)
     {
       curl_free(e_page_url);
       e_page_url = NULL;
     }
 
-  if (e_thumb_url)
+  if (e_thumb_url != NULL)
     {
       curl_free(e_thumb_url);
       e_thumb_url = NULL;
     }
 }
 
-static void dump_media_old(parsed_t p)
-{
-  spew(" > Dump media:\n"
-       "host    : %s\n"
-       "url     : %s\n"
-       "title   : %s\n"
-       "id      : %s\n"
-       "format  : %s (requested)\n",
-       p->host,
-       p->page_url,
-       p->page_title,
-       p->media_id,
-       p->format);
-
-  if (strlen(p->start_time))
-    spew("start time: %s\n", p->start_time);
-
-  if (strlen(p->thumb_url))
-    spew("thumbnail url: %s\n", p->thumb_url);
-
-  if (p->duration)
-    spew("duration: %.0f\n", p->duration);
-}
-
 static void dump_media_json(parsed_t p)
 {
   char *t;
@@ -600,13 +589,13 @@ static void dump_media_json(parsed_t p)
        p->media_id,
        p->format);
 
-  if (strlen(p->start_time))
+  if (strlen(p->start_time) >0)
     spew("  \"start_time\": \"%s\",\n", p->start_time);
 
-  if (strlen(p->thumb_url))
+  if (strlen(p->thumb_url) >0)
     spew("  \"thumbnail_url\": \"%s\",\n", p->thumb_url);
 
-  if (p->duration)
+  if (p->duration >0)
     spew("  \"duration\": \"%.0f\",\n", p->duration);
 
   spew("  \"link\": [\n");
@@ -629,134 +618,222 @@ static void dump_media(quvi_media_t media)
   quvi_getprop(media, QUVIPROP_MEDIATHUMBNAILURL, &p.thumb_url);
   quvi_getprop(media, QUVIPROP_MEDIADURATION, &p.duration);
 
-  if (opts->xml_given)
+  if (opts->xml_given == 1)
     dump_media_xml(&p);
-  else if (opts->old_given)
-    dump_media_old(&p);
   else
     dump_media_json(&p);
 
   dump_media_urls(media);
 
-  if (opts->xml_given)
+  if (opts->xml_given == 1)
     spew("</media>\n");
-  else if (opts->old_given) ;
   else
     spew("  ]\n}\n");
 }
 
-static quvi_t init_quvi()
+static void check_categories(QUVIcategory *n)
 {
+  int i;
+  for (i=0, *n=0; i<opts->category_given; ++i)
+    {
+      switch (opts->category_arg[i])
+        {
+        case category_arg_all:
+        case category__NULL:
+        default:
+          *n |= QUVIPROTO_ALL;
+          break;
+        case category_arg_http:
+          *n |= QUVIPROTO_HTTP;
+          break;
+        case category_arg_mms:
+          *n |= QUVIPROTO_MMS;
+          break;
+        case category_arg_rtmp:
+          *n |= QUVIPROTO_RTMP;
+          break;
+        case category_arg_rtsp:
+          *n |= QUVIPROTO_RTSP;
+          break;
+        }
+    }
+}
+
+static void depr_category(const char *o)
+{
+  spew_qe("warning: %s: deprecated, use --category instead\n", o);
+}
+
+static void depr_export_format(const char *o)
+{
+  spew_qe("warning: %s: deprecated, use --export-format instead\n", o);
+}
+
+static void depr_verbosity(const char *o)
+{
+  spew_qe("warning: %s: deprecated, use --verbosity instead\n", o);
+}
+
+static void depr_feat(const char *o)
+{
+  spew_qe("warning: %s: deprecated, use --feature instead\n", o);
+}
+
+static int enabled(enum enum_feature on)
+{
+  const enum enum_feature off = on+1;
+  int i,r=1;
+  for (i=0; i<opts->feature_given; ++i)
+    {
+      if (opts->feature_arg[i] == on)
+        r = 1;
+      else if (opts->feature_arg[i] == off)
+        r = 0;
+    }
+  return (r);
+}
+
+static void init_quvi()
+{
+  int resolve, verify, proxy;
+  QUVIcategory categories;
   QUVIcode rc;
-  quvi_t quvi;
 
   if ((rc = quvi_init(&quvi)) != QUVI_OK)
     {
       dump_error(quvi, rc);
-      exit(rc);
+      cleanup();
+      exit (rc);
     }
-  assert(quvi != 0);
+  assert(quvi != NULL);
 
   /* Set quvi options. */
 
-  if (opts->format_given)
-    quvi_setopt(quvi, QUVIOPT_FORMAT, opts->format_arg);
+  quvi_setopt(quvi, QUVIOPT_FORMAT, opts->format_arg);
 
-  if (opts->no_shortened_given)
-    {
-      spew_e("warning: --no-shortened is deprecated, "
-             "use --no-resolve instead\n");
-    }
-  quvi_setopt(quvi, QUVIOPT_NORESOLVE, opts->no_resolve_given);
-  quvi_setopt(quvi, QUVIOPT_NOVERIFY, opts->no_verify_given);
+  resolve = enabled(feature_arg_resolve);
+  verify  = enabled(feature_arg_verify);
 
-  if (opts->category_all_given)
-    quvi_setopt(quvi, QUVIOPT_CATEGORY, QUVIPROTO_ALL);
-  else
+  /* Deprecated. */
+
+  if (opts->no_resolve_given)
     {
-      long n = 0;
-      if (opts->category_http_given)
-        n |= QUVIPROTO_HTTP;
-      if (opts->category_mms_given)
-        n |= QUVIPROTO_MMS;
-      if (opts->category_rtsp_given)
-        n |= QUVIPROTO_RTSP;
-      if (opts->category_rtmp_given)
-        n |= QUVIPROTO_RTMP;
-      if (n > 0)
-        quvi_setopt(quvi, QUVIOPT_CATEGORY, n);
+      depr_feat("--no-resolve");
+      resolve = 0;
     }
 
+  if (opts->no_verify_given)
+    {
+      depr_feat("--no-verify");
+      verify = 0;
+    }
+
+  quvi_setopt(quvi, QUVIOPT_NORESOLVE, resolve == 0);
+  quvi_setopt(quvi, QUVIOPT_NOVERIFY, verify == 0);
+
+  /* Category. */
+
+  categories = 0;
+
+  if (opts->category_arg != NULL)
+    check_categories(&categories);
+
+  /* Category: deprecated. To be removed in later versions. */
+
+  if (opts->category_http_given == 1)
+    {
+      depr_category("--category-http");
+      categories |= QUVIPROTO_HTTP;
+    }
+  if (opts->category_rtmp_given == 1)
+    {
+      depr_category("--category-rtmp");
+      categories |= QUVIPROTO_RTMP;
+    }
+  if (opts->category_rtsp_given == 1)
+    {
+      depr_category("--category-rtsp");
+      categories |= QUVIPROTO_RTSP;
+    }
+  if (opts->category_mms_given == 1)
+    {
+      depr_category("--category-mms");
+      categories |= QUVIPROTO_MMS;
+    }
+  if (opts->category_all_given == 1)
+    {
+      depr_category("--category-all");
+      categories = QUVIPROTO_ALL;
+    }
+
+  if (categories == 0)
+    categories = QUVIPROTO_ALL;
+
+  quvi_setopt(quvi, QUVIOPT_CATEGORY, categories);
+
+  /* Status callback */
+
   quvi_setopt(quvi, QUVIOPT_STATUSFUNCTION, status_callback);
-  quvi_setopt(quvi, QUVIOPT_WRITEFUNCTION, write_callback);
 
   /* Use the quvi created cURL handle. */
 
   quvi_getinfo(quvi, QUVIINFO_CURL, &curl);
-  assert(curl != 0);
+  assert(curl != NULL);
 
-  if (opts->agent_given)
-    curl_easy_setopt(curl, CURLOPT_USERAGENT, opts->agent_arg);
+  curl_easy_setopt(curl, CURLOPT_USERAGENT, opts->agent_arg);
+  curl_easy_setopt(curl, CURLOPT_PROXY, opts->proxy_arg);
 
-  if (opts->proxy_given)
-    curl_easy_setopt(curl, CURLOPT_PROXY, opts->proxy_arg);
+  proxy = enabled(feature_arg_proxy);
 
-  if (opts->no_proxy_given)
+  /* Deprecated. */
+
+  if (opts->no_proxy_given == 1)
+    {
+      depr_feat("--no-proxy");
+      proxy = 0;
+    }
+
+#ifdef _1
+  printf("resolve=%d, verify=%d, proxy=%d\n", resolve, verify, proxy);
+#endif
+
+  if (proxy == 0)
     curl_easy_setopt(curl, CURLOPT_PROXY, "");
 
-  curl_easy_setopt(curl, CURLOPT_VERBOSE, opts->verbose_libcurl_given);
-
-  curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT,
-                   opts->connect_timeout_arg);
-
-  return (quvi);
-}
-
-static void cleanup()
-{
-  quvi_llst_free(&inputs);
-  assert(inputs == NULL);
-
-  if (quvi)
-    quvi_close(&quvi);
-  assert(quvi == NULL);
-
-  if (opts)
+  if (opts->verbose_libcurl_given)
     {
-      cmdline_parser_free(opts);
-      _free(opts);
+      depr_verbosity("--verbose-libcurl"); /* Must come before next line */
+      opts->verbosity_arg = verbosity_arg_debug;
     }
-  assert(opts == NULL);
+
+  curl_easy_setopt(curl, CURLOPT_VERBOSE,
+                   opts->verbosity_arg == verbosity_arg_debug);
+
+  curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, opts->connect_timeout_arg);
 }
 
-static void read_from(FILE *f, int close)
+static void read_from(FILE *f)
 {
   char b[256];
 
   if (!f)
     return;
 
-  while (fgets(b, sizeof(b), f))
+  while (fgets(b, (int)sizeof(b), f))
     {
-      if (strlen(b) > 16)
+      if (strlen(b) >16)
         {
           const size_t n = strlen(b)-1;
 
           if (b[n] == '\n')
             b[n] = '\0';
 
-          quvi_llst_append(&inputs, strdup(b));
+          quvi_llst_append(&input, strdup(b));
         }
     }
-
-  if (close)
-    {
-      fclose(f);
-      f = NULL;
-    }
 }
 
-static char *parse_url_scheme(const char *url)
+/*@null@*/ static char *parse_url_scheme(const char *url)
 {
   char *p, *r;
 
@@ -780,177 +857,227 @@ static int is_url(const char *s)
   return (0);
 }
 
-static FILE* open_file(const char *path)
+static void read_file(const char *path)
 {
   FILE *f = fopen(path, "rt");
-  if (!f)
+  if (f == NULL)
+    {
 #ifdef HAVE_STRERROR
-    spew_e("error: %s: %s\n", path, strerror(errno));
+      spew_e("error: %s: %s\n", path, strerror(errno));
 #else
-    perror("fopen");
+      perror("fopen");
 #endif
-  return (f);
-}
-
-static int read_input()
-{
-  if (opts->inputs_num == 0)
-    read_from(stdin, 0);
+    }
   else
     {
-      int i;
-      for (i=0; i<opts->inputs_num; ++i)
-        {
-          if (!is_url(opts->inputs[i]))
-            read_from(open_file(opts->inputs[i]), 1);
-          else /* Must be an URL. */
-            quvi_llst_append(&inputs, strdup(opts->inputs[i]));
-        }
+      read_from(f);
+      fclose(f);
+      f = NULL;
     }
-  return (quvi_llst_size(inputs));
 }
 
-int main(int argc, char *argv[])
+static unsigned int read_input()
 {
-  const char *home, *no_config, *fname;
-  QUVIcode rc, last_failure;
-  quvi_llst_node_t curr;
-  quvi_media_t media;
-  int no_config_flag;
-  int i, inputs_num;
-  int errors;
+  if (opts->inputs_num == 0)
+    read_from(stdin);
+  else
+    {
+      unsigned int i;
+      for (i=0; i<opts->inputs_num; ++i)
+        {
+          if (is_url(opts->inputs[i]) == 0)
+            read_file(opts->inputs[i]);
+          else /* Must be an URL. */
+            quvi_llst_append(&input, strdup(opts->inputs[i]));
+        }
+    }
+  return ((unsigned int)quvi_llst_size(input));
+}
 
-  assert(quvi == NULL);
-  assert(curl == NULL);
-  assert(opts == NULL);
-  assert(inputs == NULL);
+static int config_exists(const char *fpath)
+{
+  FILE *f = fopen(fpath, "r");
+  int have_config = 0;
+  if (f != NULL)
+    {
+      have_config = 1;
+      fclose(f);
+      f = NULL;
+    }
+  return (have_config);
+}
 
-  no_config = getenv("QUVI_NO_CONFIG");
-  no_config_flag = 1;
+static int parse_config(int argc, char **argv, char *home)
+{
+  struct cmdline_parser_params *pp = NULL;
+  int have_config = 0;
+  char *fpath = NULL;
 
-  home = getenv("QUVI_HOME");
-  if (!home)
-    home = getenv("HOME");
+  asprintf(&fpath, "%s/%s", home, config_fname);
+  if (fpath == NULL)
+    {
+      cleanup();
+      exit (QUVI_MEM);
+    }
 
-#ifndef HOST_W32
-  fname = "/.quvirc";
-#else
-  fname = "\\quvirc";
-#endif
+  have_config = config_exists(fpath);
+  if (have_config == 0)
+    {
+      _free(fpath);
+      return (have_config);
+    }
 
-  atexit(cleanup);
+  pp = cmdline_parser_params_create();
+  pp->check_required = 0;
+
+  have_config = (int)(cmdline_parser_config_file(fpath, opts, pp) == 0);
+  if (have_config == 1)
+    {
+      _free(fpath); /* cmdline_parser_ext may exit early. */
+
+      /* Config parsed at this point. Next, apply the cmdline options
+       * and allow overriding the values defined in the config.
+       *
+       * This step resets the $opt_given flags, e.g. if the config
+       * contains the "proxy", the parser sets proxy_given to 1,
+       * --proxy is used to override the config value, the flag
+       * is reset to 0. */
+      pp->check_required = 1;
+      pp->initialize = 0;
+      pp->override = 1;
+      have_config = (int)(cmdline_parser_ext(argc, argv, opts, pp) == 0);
+    }
+
+  _free(fpath);
+  _free(pp);
+
+  return (have_config);
+}
+
+static void parse_cmdline(int argc, char **argv)
+{
+  int have_config = 0;
+  char *home = NULL;
 
   opts = calloc(1, sizeof(struct gengetopt_args_info));
-  if (!opts)
-    return(QUVI_MEM);
+  if (opts == NULL)
+    exit (QUVI_MEM);
 
-  /* Init cmdline parser. */
+  home = getenv("QUVI_HOME");
+  if (home == NULL)
+    home = getenv("HOME");
 
-  if (home && !no_config)
-    {
-      char *path;
-      FILE *f;
+  if (home != NULL && getenv("QUVI_NO_CONFIG") == NULL)
+    have_config = parse_config(argc, argv, home);
 
-      asprintf(&path, "%s%s", home, fname);
-      f = fopen(path, "r");
-
-      if (f != NULL)
-        {
-          struct cmdline_parser_params *pp;
-
-          fclose(f);
-          f = NULL;
-
-          pp = cmdline_parser_params_create();
-          pp->check_required = 0;
-
-          if (cmdline_parser_config_file(path, opts, pp) == 0)
-            {
-              pp->initialize = 0;
-              pp->override = 1;
-              pp->check_required = 1;
-
-              if (cmdline_parser_ext(argc, argv, opts, pp) == 0)
-                no_config_flag = 0;
-            }
-          _free(pp);
-        }
-
-      _free(path);
-    }
-
-  if (no_config_flag)
+  if (have_config == 0)
     {
       if (cmdline_parser(argc, argv, opts) != 0)
-        return (QUVI_INVARG);
+        {
+          cleanup();
+          exit (QUVI_INVARG);
+        }
     }
+}
 
-  if (opts->version_given)
-    version(opts);
+static int process_queue()
+{
+  QUVIcode rc, last_error;
+  unsigned int input_num;
+  quvi_llst_node_t curr;
+  quvi_media_t media;
+  int i, errors;
 
-  if (opts->license_given)
-    license(opts);
+  input_num = read_input();
+  last_error = QUVI_OK;
+  errors = 0;
 
-  verbose_flag = !opts->quiet_given;
-
-  quvi = init_quvi();
-
-  if (opts->query_formats_given)
-    query_formats(quvi);
-
-  if (opts->support_given)
-    support(quvi);
-
-  if (opts->format_given)
-    format_help(quvi);
-
-  /* User input */
-
-  inputs_num = read_input();
-
-  if (inputs_num == 0)
+  if (input_num == 0)
     {
       spew_qe("error: no input URLs\n");
       return (QUVI_INVARG);
     }
 
-  last_failure = QUVI_OK;
-  errors = 0;
-
-  for (i=0, curr=inputs; curr; ++i)
+  for (i=0, curr=input; curr; ++i)
     {
       char *url = quvi_llst_data(curr);
       rc = quvi_parse(quvi, url, &media);
       if (rc == QUVI_OK)
         {
-          assert(media != 0);
+          int i;
+
+          assert(media != NULL);
           dump_media(media);
 
-          if (opts->exec_given)
+          for (i=0; i<opts->exec_given; ++i)
             {
               do
                 {
-                  invoke_exec(media);
+                  invoke_exec(media, opts->exec_arg[i]);
                 }
               while (quvi_next_media_url(media) == QUVI_OK);
             }
         }
       else
         {
-          dump_error(quvi,rc);
-          last_failure = rc;
+          dump_error(quvi, rc);
+          last_error = rc;
           ++errors;
         }
       quvi_parse_close(&media);
-      assert(media == 0);
+      assert(media == NULL);
       curr = quvi_llst_next(curr);
     }
 
-  if (inputs_num > 1)
+  if (input_num >1)
     {
       spew_qe("Results: %d OK, %d failed (last 0x%02x), exit with 0x%02x\n",
-              inputs_num - errors, errors, last_failure, rc);
+              input_num - errors, errors, last_error, rc);
     }
+  return (rc);
+}
+
+int main(int argc, char *argv[])
+{
+  int rc;
+
+  assert(quvi == NULL);
+  assert(curl == NULL);
+  assert(opts == NULL);
+  assert(input == NULL);
+
+  parse_cmdline(argc, argv);
+
+  if (opts->version_given == 1)
+    print_version();
+
+  if (opts->license_given == 1)
+    license();
+
+  init_quvi();
+
+  /* Deprecated. */
+
+  if (opts->xml_given)
+    {
+      opts->export_format_arg = export_format_arg_xml;
+      depr_export_format("--xml");
+    }
+
+  if (opts->quiet_given == 1)
+    {
+      depr_verbosity("--quiet"); /* Must come before the next line */
+      opts->verbosity_arg = verbosity_arg_quiet;
+    }
+
+  if (opts->query_formats_given == 1)
+    query_formats(quvi);
+
+  if (opts->support_given == 1)
+    support(quvi);
+
+  rc = process_queue();
+  cleanup();
 
   return (rc);
 }
